@@ -48,12 +48,6 @@ macro_rules! pattern {
         $crate::macros::TestHelper::<Pattern>::from($arg).0
     };
 }
-#[macro_export]
-macro_rules! ptn {
-    ($arg:expr) => {
-        $crate::macros::TestHelper::<Term>::from(pattern!($arg)).0
-    };
-}
 
 #[macro_export]
 macro_rules! param {
@@ -78,22 +72,6 @@ macro_rules! instance {
     };
 }
 
-#[macro_export]
-macro_rules! partial {
-    ($arg:expr) => {
-        Value::Partial(Partial::new(sym!($arg)))
-    };
-    ($arg:expr, [$($args:expr),*]) => {
-        {
-            let mut constraint = Partial::new(sym!($arg));
-            $(
-                constraint.add_constraint($args);
-            )*
-            constraint
-        }
-    };
-}
-
 #[macro_export]
 macro_rules! sym {
     ($arg:expr) => {
@@ -171,16 +149,6 @@ macro_rules! op {
     };
 }
 
-#[macro_export]
-macro_rules! opn {
-    ($op_type:ident, $($args:expr),+) => {
-        op!($op_type, $($args),+).into_term()
-    };
-    ($op_type:ident) => {
-        op!($op_type).into_term()
-    };
-}
-
 #[macro_export]
 macro_rules! dict {
     ($arg:expr) => {
@@ -220,6 +188,7 @@ macro_rules! rule {
             params,
             body: term!(op!(And, $(term!($body)),+)),
             source_info: $crate::sources::SourceInfo::Test,
+            required: false,
         }}
     };
     ($name:expr, [$($args:tt)*]) => {{
@@ -230,6 +199,20 @@ macro_rules! rule {
             params,
             body: term!(op!(And)),
             source_info: $crate::sources::SourceInfo::Test,
+            required: false,
+        }
+    }};
+    // this macro variant is used exclusively to create rule *types*
+    // TODO: @patrickod break into specific-purpose rule_type! macro and RuleType struct
+    ($name:expr, [$($args:tt)*], $required:expr) => {{
+        let mut params = args!($($args)*);
+        params.reverse();
+        Rule {
+            name: sym!($name),
+            params,
+            body: term!(op!(And)),
+            source_info: $crate::sources::SourceInfo::Test,
+            required: $required,
         }
     }};
 }
@@ -247,7 +230,17 @@ impl<T> From<T> for TestHelper<T> {
 
 impl From<Value> for TestHelper<Term> {
     fn from(other: Value) -> Self {
-        Self(Term::new_temporary(other))
+        Self(Term::from(other))
+    }
+}
+
+impl From<u64> for ExternalInstance {
+    fn from(instance_id: u64) -> Self {
+        ExternalInstance {
+            instance_id,
+            constructor: None,
+            repr: None,
+        }
     }
 }
 
@@ -272,7 +265,7 @@ impl From<Value> for TestHelper<Parameter> {
     /// a specializer.
     fn from(name: Value) -> Self {
         Self(Parameter {
-            parameter: Term::new_temporary(name),
+            parameter: Term::from(name),
             specializer: None,
         })
     }
@@ -367,6 +360,6 @@ impl From<InstanceLiteral> for TestHelper<Pattern> {
 }
 impl From<Pattern> for TestHelper<Term> {
     fn from(other: Pattern) -> Self {
-        Self(Term::new_temporary(value!(other)))
+        Self(Term::from(value!(other)))
     }
 }
