@@ -1,12 +1,14 @@
+use std::sync::{Arc, RwLock};
+
 use super::data_filtering::{build_filter_plan, FilterPlan, PartialResults, Types};
-use super::diagnostic::Diagnostic;
+use super::diagnostic::{set_context_for_diagnostics, Diagnostic};
 use super::error::PolarResult;
-use super::events::*;
 use super::kb::*;
 use super::messages::*;
 use super::parser;
+use super::query::Query;
+use super::resource_block::resource_block_from_productions;
 use super::rewrites::*;
-use super::runnable::Runnable;
 use super::sources::*;
 use super::terms::*;
 use super::validations::{
@@ -15,129 +17,6 @@ use super::validations::{
 };
 use super::vm::*;
 
-use std::sync::{Arc, RwLock};
-
-pub struct Query {
-    runnable_stack: Vec<(Box<dyn Runnable>, u64)>, // Tuple of Runnable + call_id.
-    vm: PolarVirtualMachine,
-    term: Term,
-    done: bool,
-}
-
-impl Query {
-    pub fn new(vm: PolarVirtualMachine, term: Term) -> Self {
-        Self {
-            runnable_stack: vec![],
-            vm,
-            term,
-            done: false,
-        }
-    }
-
-    #[cfg(target_arch = "wasm32")]
-    pub fn set_logging_options(&mut self, rust_log: Option<String>, polar_log: Option<String>) {
-        self.vm.set_logging_options(rust_log, polar_log);
-    }
-
-    /// Runnable lifecycle
-    ///
-    /// 1. Get Runnable A from the top of the Runnable stack, defaulting to the VM.
-    /// 2. If Runnable A emits a Run event containing Runnable B, push Runnable B onto the stack.
-    /// 3. Immediately request the next event, which will execute Runnable B.
-    /// 4. When Runnable B emits a Done event, pop Runnable B off the stack and return its result as
-    ///    an answer to Runnable A.
-    pub fn next_event(&mut self) -> PolarResult<QueryEvent> {
-        let mut counter = self.vm.id_counter();
-        let qe = match self.top_runnable().run(Some(&mut counter)) {
-            Ok(e) => e,
-            Err(e) => self.top_runnable().handle_error(e)?,
-        };
-        self.recv_event(qe)
-    }
-
-    fn recv_event(&mut self, qe: QueryEvent) -> PolarResult<QueryEvent> {
-        match qe {
-            QueryEvent::None => self.next_event(),
-            QueryEvent::Run { runnable, call_id } => {
-                self.push_runnable(runnable, call_id);
-                self.next_event()
-            }
-            QueryEvent::Done { result } => {
-                if let Some((_, result_call_id)) = self.pop_runnable() {
-                    self.top_runnable()
-                        .external_question_result(result_call_id, result)?;
-                    self.next_event()
-                } else {
-                    // VM is done.
-                    assert!(self.runnable_stack.is_empty());
-                    Ok(QueryEvent::Done { result })
-                }
-            }
-            ev => Ok(ev),
-        }
-    }
-
-    fn top_runnable(&mut self) -> &mut (dyn Runnable) {
-        self.runnable_stack
-            .last_mut()
-            .map(|b| b.0.as_mut())
-            .unwrap_or(&mut self.vm)
-    }
-
-    fn push_runnable(&mut self, runnable: Box<dyn Runnable>, call_id: u64) {
-        self.runnable_stack.push((runnable, call_id));
-    }
-
-    fn pop_runnable(&mut self) -> Option<(Box<dyn Runnable>, u64)> {
-        self.runnable_stack.pop()
-    }
-
-    pub fn call_result(&mut self, call_id: u64, value: Option<Term>) -> PolarResult<()> {
-        self.top_runnable().external_call_result(call_id, value)
-    }
-
-    pub fn question_result(&mut self, call_id: u64, result: bool) -> PolarResult<()> {
-        self.top_runnable()
-            .external_question_result(call_id, result)
-    }
-
-    pub fn application_error(&mut self, message: String) -> PolarResult<()> {
-        self.vm.external_error(message)
-    }
-
-    pub fn debug_command(&mut self, command: &str) -> PolarResult<()> {
-        self.top_runnable().debug_command(command)
-    }
-
-    pub fn next_message(&self) -> Option<Message> {
-        self.vm.messages.next()
-    }
-
-    pub fn source_info(&self) -> String {
-        self.vm.term_source(&self.term, true)
-    }
-
-    pub fn bind(&mut self, name: Symbol, value: Term) -> PolarResult<()> {
-        self.vm.bind(&name, value)
-    }
-}
-
-// Query as an iterator returns `None` after the first time `Done` is seen
-impl Iterator for Query {
-    type Item = PolarResult<QueryEvent>;
-
-    fn next(&mut self) -> Option<PolarResult<QueryEvent>> {
-        if self.done {
-            return None;
-        }
-        let event = self.next_event();
-        if let Ok(QueryEvent::Done { .. }) = event {
-            self.done = true;
-        }
-        Some(event)
-    }
-}
-
 pub struct Polar {
     pub kb: Arc<RwLock<KnowledgeBase>>,
     messages: MessageQueue,
@@ -216,9 +95,16 @@ impl Polar {
                             kb.add_rule_type(rule_type);
                         }
                     }
-                    parser::Line::ResourceBlock(block) => {
-                        diagnostics.append(&mut block.add_to_kb(kb))
-                    }
+                    parser::Line::ResourceBlock {
+                        keyword,
+                        resource,
+                        productions,
+                    } => match resource_block_from_productions(keyword, resource, productions)
+                        .map(|block| block.add_to_kb(kb))
+                    {
+                        Ok(errors) | Err(errors) => diagnostics
+                            .append(&mut errors.into_iter().map(Diagnostic::Error).collect()),
+                    },
                 }
             }
             Ok(diagnostics)
@@ -237,7 +123,13 @@ impl Polar {
         }
 
         // Rewrite shorthand rules in resource blocks before validating rule types.
-        diagnostics.append(&mut kb.rewrite_shorthand_rules());
+        diagnostics.append(
+            &mut kb
+                .rewrite_shorthand_rules()
+                .into_iter()
+                .map(Diagnostic::Error)
+                .collect(),
+        );
 
         // TODO(gj): need to bomb out before rule type validation in case additional rule types
         // were defined later on in the file that encountered the `ParseError`. Those additional
@@ -245,11 +137,17 @@ impl Polar {
         // well-parsed file that also contains rules that don't conform to the shapes laid out in
         // the well-parsed file but *would have* conformed to the shapes laid out in the file that
         // failed to parse.
-        if diagnostics.iter().any(Diagnostic::is_error) {
+        if diagnostics.iter().any(Diagnostic::is_parse_error) {
+            // NOTE(gj): need to set context _before_ clearing the KB so we still have source info.
+            set_context_for_diagnostics(&kb, &mut diagnostics);
             kb.clear_rules();
             return diagnostics;
         }
 
+        // Generate appropriate rule_type definitions using the types contained in policy resource
+        // blocks.
+        kb.create_resource_specific_rule_types();
+
         // check rules are valid against rule types
         diagnostics.append(&mut kb.validate_rules());
 
@@ -265,6 +163,9 @@ impl Polar {
             diagnostics.push(w)
         };
 
+        // NOTE(gj): need to set context _before_ clearing the KB so we still have source info.
+        set_context_for_diagnostics(&kb, &mut diagnostics);
+
         // If we've encountered any errors, clear the KB.
         if diagnostics.iter().any(Diagnostic::is_error) {
             kb.clear_rules();
@@ -350,7 +251,7 @@ impl Polar {
     }
 
     pub fn register_constant(&self, name: Symbol, value: Term) -> PolarResult<()> {
-        self.kb.write().unwrap().constant(name, value)
+        self.kb.write().unwrap().register_constant(name, value)
     }
 
     /// Register MRO for `name` with `mro`.
@@ -463,7 +364,7 @@ mod tests {
         let next = diagnostics.next().unwrap();
         assert!(matches!(next, Diagnostic::Error(_)));
         assert!(
-            next.to_string().starts_with("Call to undefined rule \"g\""),
+            next.to_string().starts_with("Call to undefined rule: g()"),
             "{}",
             next
         );
